/*
  This file is part of CinLife

  CinLife is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  CinLife is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with CinLife.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

#include "mpi.h"
#include "CinLife.h"

/* Options to be parsed */
static const char*         options = "x:y:n:h";
static const struct option long_options[] = {
  {"xsize",required_argument,NULL,'x'},
  {"ysize",required_argument,NULL,'y'},
  {"nstep",required_argument,NULL,'n'}  
};

/* Static functions */
static int  cart2linear(int ii, int jj, int ncols);
static void usage();

/* Debug macros */
#ifndef DRANK
#define DRANK 0
#endif

void initialize_life(local_life* life) {
    
  /* Get rank and size of MPI_COMM_WORLD */
  MPI_Comm_rank(MPI_COMM_WORLD,&life->rank);
  MPI_Comm_size(MPI_COMM_WORLD,&life->size);
  /* MPI_Dims_create modifies ONLY ZERO ENTRIES */
  life->cartesian_communicator_length[0] = life->cartesian_communicator_length[1] = 0;
  MPI_Dims_create(life->size,TOPOLOGY_DIMENSION,
			life->cartesian_communicator_length);
  /* Create the cartesian topology */
  int is_periodic[2];
  is_periodic[0] = is_periodic[1] = true;
  MPI_Cart_create(MPI_COMM_WORLD,TOPOLOGY_DIMENSION,
		  life->cartesian_communicator_length,
		  is_periodic,true,&life->cartesian_communicator);
  /* Rank in the cartesian communicator */
  MPI_Comm_rank(life->cartesian_communicator,&life->cartesian_communicator_rank);
  MPI_Comm_size(life->cartesian_communicator,&life->cartesian_communicator_size);
  /* Cartesian coordinates */
  MPI_Cart_coords(life->cartesian_communicator,
		  life->cartesian_communicator_rank,
		  TOPOLOGY_DIMENSION,life->cartesian_coords);
  /* Compute the dimension of local regions */
  life->ncols     = life->gsizex/life->cartesian_communicator_length[0]; 
  life->nrows     = life->gsizey/life->cartesian_communicator_length[1];
  life->extra_row = false;
  life->extra_col = false;
  if (life->cartesian_coords[0] < (life->gsizex%life->cartesian_communicator_length[0])) {
    ++life->ncols;
    life->extra_col = true;
  }
  if (life->cartesian_coords[1] < (life->gsizey%life->cartesian_communicator_length[1])) {
    ++life->nrows;
    life->extra_row = true;
  }
  /* Allocate the local regions and set them to UNDEF */
  int local_size = (life->nrows +2)*(life->ncols + 2);
  life->count        = malloc( (life->nrows*life->ncols)*sizeof(char) );
  life->local_domain = malloc( local_size*sizeof(char) );
  for (int ii = 0; ii < local_size; ++ii)
    life->local_domain[ii] = UNDEF;
  /* Allocate the global region on rank 0 processor, set to UNDEF */
  if ( life->rank == 0 ) {
    int global_size = life->gsizex*life->gsizey;
    life->global_domain = malloc( global_size*sizeof(char) );
    for (int ii = 0; ii < global_size; ++ii)
      life->global_domain[ii] = UNDEF;
  } else {
    life->global_domain = NULL;
  }

  /* 
     Definition of MPI Datatypes:
     - mapping of local domains to global domain
     - local domains (excluding halo regions)
     - local columns (for boundary swaps)
     - local rows    (for boundary swaps)
  */

  /* Compute quantities for the base case: no extra column, no extra row */
  int nrows = life->gsizey/life->cartesian_communicator_length[1];
  int ncols = life->gsizex/life->cartesian_communicator_length[0];

  /* Mapping from local domains to global domain */
  /* Standard */
  MPI_Type_vector(nrows  ,ncols  ,life->gsizex,MPI_CHAR,&life->global_mapping_types[0][0]);
  MPI_Type_commit(&life->global_mapping_types[0][0]);
  /* Extra row */
  MPI_Type_vector(nrows+1,ncols  ,life->gsizex,MPI_CHAR,&life->global_mapping_types[1][0]);
  MPI_Type_commit(&life->global_mapping_types[1][0]);
  /* Extra column */
  MPI_Type_vector(nrows  ,ncols+1,life->gsizex,MPI_CHAR,&life->global_mapping_types[0][1]);
  MPI_Type_commit(&life->global_mapping_types[0][1]);
  /* Extra row and extra column */
  MPI_Type_vector(nrows+1,ncols+1,life->gsizex,MPI_CHAR,&life->global_mapping_types[1][1]);
  MPI_Type_commit(&life->global_mapping_types[1][1]);

  /* Local domains without halo layers */
  MPI_Type_vector(life->nrows,life->ncols,life->ncols + 2,MPI_CHAR,&life->local_domain_type);
  MPI_Type_commit(&life->local_domain_type);
  /* Column type on local domains */
  MPI_Type_vector(life->nrows,1,life->ncols + 2,MPI_CHAR,&life->local_column_type);
  MPI_Type_commit(&life->local_column_type);
  /* Row type on local domains */
  MPI_Type_vector(1,life->ncols + 2,life->ncols + 2,MPI_CHAR,&life->local_row_type);
  MPI_Type_commit(&life->local_row_type);
}

void set_initial_condition(local_life* life) {
  
  /* Set element in local array to white or black depending on process_coords */
  int color;
  if ( (life->cartesian_coords[0] + life->cartesian_coords[1])%2 == 0 ) 
    color = DEAD;
  else
    color = ALIVE;

  for (int ii = 0; ii < life->nrows; ++ii)
    for (int jj = 0; jj < life->ncols; ++jj) {
      life->local_domain[cart2linear(ii,jj,life->ncols)] = color;
    }
}

void swap_local_boundary(local_life* life) {

  int  neighbor_coords[2];
  int  left,right,above,bottom = -1;

  MPI_Request request[2];
  MPI_Status  status [2];
  
  /* Left neighbor */
  neighbor_coords[0] = life->cartesian_coords[0] - 1;
  neighbor_coords[1] = life->cartesian_coords[1];
  MPI_Cart_rank(life->cartesian_communicator,neighbor_coords,&left);
  /* Right neighbor */
  neighbor_coords[0] = life->cartesian_coords[0] + 1;
  neighbor_coords[1] = life->cartesian_coords[1];
  MPI_Cart_rank(life->cartesian_communicator,neighbor_coords,&right);
  /* Above neighbor */
  neighbor_coords[0] = life->cartesian_coords[0];
  neighbor_coords[1] = life->cartesian_coords[1] - 1;
  MPI_Cart_rank(life->cartesian_communicator,neighbor_coords,&above);
  /* Bottom neighbor */
  neighbor_coords[0] = life->cartesian_coords[0];
  neighbor_coords[1] = life->cartesian_coords[1] + 1;
  MPI_Cart_rank(life->cartesian_communicator,neighbor_coords,&bottom);

#ifdef DEBUGRANK0
  if (life->rank == DRANK) {
    printf("Before swap: rank %d\n",life->rank);
    printf("%dX%d local space\n",life->nrows,life->ncols);
    printf("Start send type: %d %d\n",rbool, cbool);
    printf("ISEND parameters:\n");
    printf("Lstart: %d\n",cart2linear(0,0,life->ncols));
    printf("Rstart: %d\n",cart2linear(0,life->ncols-1,life->ncols));
    printf("Astart: %d\n",cart2linear(0,-1,life->ncols));
    printf("Bstart: %d\n",cart2linear(life->nrows-1,-1,life->ncols));
    printf("\n");
    print_local_region(life);
    printf("\n");
  }
#endif

  /* Send columns to left and right neighbors */
  MPI_Isend(&life->local_domain[cart2linear(0,0,life->ncols)],1,
	    life->local_column_type,left,10,
	    life->cartesian_communicator,&request[0]);
  MPI_Isend(&life->local_domain[cart2linear(0,life->ncols-1,life->ncols)],1,
	    life->local_column_type,right,10,
	    life->cartesian_communicator,&request[1]);
  /* Receive columns from left and right neighbors */
  MPI_Recv (&life->local_domain[cart2linear(0,-1,life->ncols)],1,
	    life->local_column_type,left,10,
	    life->cartesian_communicator,&status[0]);
  MPI_Recv (&life->local_domain[cart2linear(0,life->ncols,life->ncols)],1,
	    life->local_column_type,right,10,
	    life->cartesian_communicator,&status[1]);
  /* Wait for communications to end */
  MPI_Wait(&request[0],&status[0]);
  MPI_Wait(&request[1],&status[1]);

#ifdef DEBUGRANK0
  if (life->rank == DRANK) {
    printf("After communication with left/right: rank %d\n",life->rank);
    printf("Left  : %d\n",left);
    printf("Right : %d\n",right);   
    print_local_region(life);
    printf("\n");
  }
#endif

  /* Send columns to above and bottom neighbors */
  MPI_Isend(&life->local_domain[cart2linear(0,-1,life->ncols)],1,
	    life->local_row_type,above,20,
	    life->cartesian_communicator,&request[0]);
  MPI_Isend(&life->local_domain[cart2linear(life->nrows-1,-1,life->ncols)],1,
	    life->local_row_type,bottom,20,
	    life->cartesian_communicator,&request[1]);
  /* Receive columns from above and bottom neighbors */
  MPI_Recv (&life->local_domain[cart2linear(-1,-1,life->ncols)],1,
	    life->local_row_type,above,20,
	    life->cartesian_communicator,&status[0]);
  MPI_Recv (&life->local_domain[cart2linear(life->nrows,-1,life->ncols)],1,
	    life->local_row_type,bottom,20,
	    life->cartesian_communicator,&status[1]);
  /* Wait for communications to end */
  MPI_Wait(&request[0],&status[0]);
  MPI_Wait(&request[1],&status[1]);

#ifdef DEBUGRANK0
  if (life->rank == DRANK) {
    printf("After communication with above/bottom: rank %d\n",life->rank);
    printf("Above  : %d\n",above);
    printf("Bottom : %d\n",bottom);
    print_local_region(life);
    printf("\n");
  }
#endif

  /* Reset boundaries if it is the case */
  if (life->cartesian_coords[0] == 0) {
    for (int ii = -1; ii < life->nrows + 1; ++ii){
      life->local_domain[cart2linear(ii,-1,life->ncols)] = OUTOFBORDER;
    }
  }
  if (life->cartesian_coords[0] == (life->cartesian_communicator_length[0] - 1) ) {
    for (int ii = -1; ii < life->nrows + 1; ++ii){
      life->local_domain[cart2linear(ii,life->ncols,life->ncols)] = OUTOFBORDER;
    }
  }
  if (life->cartesian_coords[1] == 0) {
    for (int jj = -1; jj < life->ncols + 1; ++jj){
      life->local_domain[cart2linear(-1,jj,life->ncols)] = OUTOFBORDER;
    }
  }
  if (life->cartesian_coords[1] == (life->cartesian_communicator_length[1] - 1) ) {
    for (int jj = -1; jj < life->ncols + 1; ++jj){
      life->local_domain[cart2linear(life->nrows,jj,life->ncols)] = OUTOFBORDER;
    }
  }
    
#ifdef DEBUGRANK0
  if (life->rank == DRANK) {
    printf("After resetting external borders: rank %d\n",life->rank);
    print_local_region(life);
    printf("\n");
  }
#endif

}

void update_local_domain (local_life* life) {

  /* Count number of alive neighbors */
  for (int ii = 0; ii < life->nrows; ++ii )
    for(int jj = 0; jj < life->ncols; ++jj) {
      int alive = 0;
      /* Count how many neighbors are alive */
      for (int nridx = ii - 1; nridx < ii + 2; ++nridx)
	for (int ncidx = jj - 1; ncidx < jj +2; ++ncidx) {
	  if ( life->local_domain[cart2linear(nridx,ncidx,life->ncols)] == ALIVE )
	    ++alive;
	}
      if ( life->local_domain[cart2linear(ii,jj,life->ncols)] == ALIVE )
	--alive;
      life->count[ii*(life->ncols) + jj] = alive;
    }
  /* Update local domain */
  for (int ii = 0; ii < life->nrows; ++ii )
    for(int jj = 0; jj < life->ncols; ++jj) {
      if ( ( life->count[ii*life->ncols + jj] < 2 ) || ( life->count[ii*life->ncols + jj] > 3 ) )
	life->local_domain[cart2linear(ii,jj,life->ncols)] = DEAD;
      else if (life->count[ii*life->ncols + jj] == 3 )
	life->local_domain[cart2linear(ii,jj,life->ncols)] = ALIVE;
    }
}

void update_global_domain(local_life* life) {

  if ( life->cartesian_communicator_rank == 0 ) {    
    int        has_extra_row;
    int        has_extra_col;
    int        sender_coords[2];
    int        startidx;
    int        stridecol,striderow;
    MPI_Status status;

    int nrows = life->gsizey/life->cartesian_communicator_length[1];
    int ncols = life->gsizex/life->cartesian_communicator_length[0];

    /* Copy local buffer */
    for(int ii = 0; ii < life->nrows; ++ii) {
      for(int jj = 0; jj < life->ncols; ++jj) {
	life->global_domain[life->gsizex*ii + jj] = life->local_domain[cart2linear(ii,jj,life->ncols)];
      }
    }
    /* Receive other task buffer */
    for(int ii = 1; ii < life->cartesian_communicator_size; ++ii) {
      MPI_Cart_coords(life->cartesian_communicator,ii,TOPOLOGY_DIMENSION,sender_coords);
      has_extra_row = sender_coords[1] < (life->gsizey%life->cartesian_communicator_length[1]);
      has_extra_col = sender_coords[0] < (life->gsizex%life->cartesian_communicator_length[0]);

      /* Stride on columns */
      if (has_extra_col == 0) 
	stridecol = life->gsizex%life->cartesian_communicator_length[0];
      else
	stridecol = sender_coords[0];
      /* Stride on rows */
      if (has_extra_row == 0)
	striderow = life->gsizey%life->cartesian_communicator_length[1];
      else
	striderow = sender_coords[1];
      /* Set startidx at the beginning of the row in the topology */
      startidx  = (sender_coords[1]*nrows + striderow)*life->gsizex;
      startidx +=  sender_coords[0]*ncols + stridecol;
      MPI_Recv(&life->global_domain[startidx],1,life->global_mapping_types[has_extra_row][has_extra_col],
	       ii,10,life->cartesian_communicator,&status);            
      
    }
  } else {
    MPI_Send(&life->local_domain[cart2linear(0,0,life->ncols)],1,
	     life->local_domain_type,0,10,life->cartesian_communicator);
  }
}

void print_local_count(local_life* life) {
  printf("\n");
  for (int ii = 0; ii < life->nrows; ++ii) {
    printf("\n");
    for(int jj = 0; jj < life->ncols; ++jj)
      printf("%3.1d ",life->count[ii*(life->ncols) + jj]);      
  }
  printf("\n\n");
}

void print_to_stdout(const local_life* life) {

  if (life->cartesian_communicator_rank == 0) {
    /* Print buffer for debug purpose */
    for (int ii = 0; ii < life->gsizex*life->gsizey; ++ii) {
      if (ii%life->gsizex == 0)
	printf("\n");
      printf("%3.1d ",life->global_domain[ii]);
    }
    printf("\n\n");
  }
}

void terminate_life(local_life* life) {
  
  /* Free local_storage */
  if (life->local_domain)
    free(life->local_domain);
  if (life->global_domain)
    free(life->global_domain);
  if (life->count)
    free(life->count);
  
  /* Free datatypes */
  for(int ii = 0; ii < 2; ++ii)
    for (int jj = 0; jj < 2; ++jj) {
      MPI_Type_free(&life->global_mapping_types[ii][jj]);
    }
  MPI_Type_free(&life->local_domain_type);
  MPI_Type_free(&life->local_column_type);
  MPI_Type_free(&life->local_row_type   );
  
  /* Free communicators */
  MPI_Comm_free(&life->cartesian_communicator);
}

void print_local_region(const local_life* life) {
  printf("\n");
  for (int ii = 0; ii < life->nrows + 2; ++ii) {
    printf("\n");
    for(int jj = 0; jj < life->ncols + 2; ++jj)
      printf("%3.1d ",life->local_domain[ii*(life->ncols + 2) + jj]);      
  }
  printf("\n\n");
}

void read_input_data(int argc,char** argv,local_life* life){

  int opt = 0;
  int idx = 0;
  
  /* Set defaults */
  life->gsizex = DEFAULT_SIZE_X;
  life->gsizey = DEFAULT_SIZE_Y;
  life->nstep  = DEFAULT_NSTEP;
  /* Get options from command line */
  for(;;){

    opt = getopt_long(argc,argv,options,long_options,&idx);
    if (opt == -1) break;

    switch(opt) {
    case 'x':
      life->gsizex = strtol(optarg,NULL,10);
      break;
    case 'y':
      life->gsizey = strtol(optarg,NULL,10);
      break;
    case 'n':
      life->nstep = strtol(optarg,NULL,10);
      break;
    case 'h':
      usage();
      break;
    default:
      break;
    }
  }
}

void print_to_pgm_file(const char* filename, const local_life* life) {

  FILE * outfile;
  if (life->rank == 0) {
    /* Open file */
    outfile = fopen(filename,"wb");
    /* Write header */
    fprintf(outfile,"P5\n");
    /* Write width, height, maximum gray value */
    fprintf(outfile,"%d %d\n%d\n",life->gsizex,life->gsizey,ALIVE);
    /* Write the life board */
    fwrite(life->global_domain,sizeof(char),life->gsizex*life->gsizey,outfile);
    /* Close file */
    fclose(outfile);
  }

}

static void usage() {
  int rank = 0;
  MPI_Comm_rank(MPI_COMM_WORLD,&rank) ;  
  if (rank == 0) {
    printf("\nmpirun -np [nprocs] ./GameOfLife -x [xsize] -y [ysize] -n [nstep] \n");
    printf("\n");
    printf("-x [xsize] : set the size of the board along x-axis (default: %d)\n",DEFAULT_SIZE_X);
    printf("-y [ysize] : set the size of the board along y-axis (default: %d)\n",DEFAULT_SIZE_Y);
    printf("-n [nstep] : set the number of step to be simulated (default: %d)\n",DEFAULT_NSTEP);
    printf("\n");
    printf("Use your favourite viewer to check the output files! (gqview,xv)\n\n");
  }
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  exit(EXIT_FAILURE);      
}

static int cart2linear(int ii, int jj, int ncols) {
  return (jj + 1) + (ii + 1)*(ncols + 2);
}

